class TablePathFinder:
    def __init__(self, content: str) -> None:
        self.lines = content.splitlines()

    def find_line(self, target_path: list[str]) -> int:
        target_key = target_path[-1]
        target_parents = target_path[:-1]
        current_section: list[str] = []

        for i, line in enumerate(self.lines, 1):
            stripped = line.strip()
            if not stripped or stripped.startswith(("#", ";")):
                continue

            # Continuation line (starts with whitespace)
            if line[0] in (" ", "\t"):
                continue

            # Working with sections [section] or [parent.child]
            if stripped.startswith("[") and "]" in stripped:
                section_name = stripped.strip("[] ")
                current_section = section_name.split(".")
                continue

            # Working with keys key = value
            sep = "=" if "=" in stripped else ":"
            if sep in stripped:
                key = stripped.partition(sep)[0].strip().strip("\"'")
                if key == target_key and current_section == target_parents:
                    return i
        return -1



from dataclasses import dataclass


@dataclass
class _StackEntry:
    key: str | None
    is_array: bool
    array_index: int


@dataclass
class _ParseState:
    pos: int
    line: int
    last_key: str | None


@dataclass(frozen=True, slots=True)
class _PosLine:
    pos: int
    line: int


@dataclass(frozen=True, slots=True)
class _KeyResult:
    pos: int
    line: int
    last_key: str | None
    found: bool


class Json5PathFinder:
    def __init__(self, content: str) -> None:
        self._content = content

    def find_line(self, target_path: list[str]) -> int:
        content = self._content
        length = len(content)
        state = _ParseState(pos=0, line=1, last_key=None)
        stack: list[_StackEntry] = []

        while state.pos < length:
            ch = content[state.pos]

            if _try_skip_noise(ch, content, state, length):
                continue

            if ch in "{[":
                _increment_parent_array(stack)
                is_array = ch == "["
                stack.append(_StackEntry(key=state.last_key, is_array=is_array, array_index=-1))
                state.pos += 1
                state.last_key = None
                continue

            if ch in "}]":
                if stack:
                    stack.pop()
                state.pos += 1
                state.last_key = None
                continue

            if ch in {'"', "'"}:
                result = _handle_quoted(content, state, length, stack, target_path)
                if result.found:
                    return result.line
                state.pos = result.pos
                state.line = result.line
                state.last_key = result.last_key
                continue

            if _is_ident_start(ch):
                result = _handle_identifier(content, state, length, stack, target_path)
                if result.found:
                    return result.line
                state.pos = result.pos
                state.line = result.line
                state.last_key = result.last_key
                continue

            _increment_parent_array(stack)
            state.pos = _skip_scalar(content, state.pos, length)

        return -1


def _try_skip_noise(ch: str, content: str, state: _ParseState, length: int) -> bool:
    if ch in " \t\r\n":
        if ch == "\n":
            state.line += 1
        state.pos += 1
        return True

    if ch == ",":
        state.pos += 1
        return True

    if ch != "/" or state.pos + 1 >= length:
        return False

    next_ch = content[state.pos + 1]
    if next_ch == "/":
        state.pos = _skip_line_comment(content, state.pos + 2, length)
        return True
    if next_ch == "*":
        block = _skip_block_comment(content, state.pos + 2, length, state.line)
        state.pos = block.pos
        state.line = block.line
        return True

    return False


def _handle_quoted(
    content: str,
    state: _ParseState,
    length: int,
    stack: list[_StackEntry],
    target_path: list[str],
) -> _KeyResult:
    quote = content[state.pos]
    string_start = state.pos + 1
    end = _skip_quoted_string(content, string_start, quote, state.line)
    string_value = content[string_start : end.pos - 1]

    colon = _skip_ws_and_comments(content, end.pos, length, end.line)

    if colon.pos >= length or content[colon.pos] != ":":
        _increment_parent_array(stack)
        return _KeyResult(pos=end.pos, line=end.line, last_key=None, found=False)

    if _build_path(stack, string_value) == target_path:
        return _KeyResult(pos=end.pos, line=state.line, last_key=None, found=True)

    return _KeyResult(pos=colon.pos + 1, line=colon.line, last_key=string_value, found=False)


def _handle_identifier(
    content: str,
    state: _ParseState,
    length: int,
    stack: list[_StackEntry],
    target_path: list[str],
) -> _KeyResult:
    ident_end = _skip_identifier(content, state.pos, length)
    ident = content[state.pos : ident_end]

    colon = _skip_ws_and_comments(content, ident_end, length, state.line)

    if colon.pos >= length or content[colon.pos] != ":":
        _increment_parent_array(stack)
        return _KeyResult(pos=ident_end, line=state.line, last_key=None, found=False)

    if _build_path(stack, ident) == target_path:
        return _KeyResult(pos=ident_end, line=state.line, last_key=None, found=True)

    return _KeyResult(pos=colon.pos + 1, line=colon.line, last_key=ident, found=False)


def _increment_parent_array(stack: list[_StackEntry]) -> None:
    if stack and stack[-1].is_array:
        stack[-1].array_index += 1


def _build_path(stack: list[_StackEntry], current_key: str) -> list[str]:
    path: list[str] = []
    for entry in stack:
        if entry.key is not None:
            path.append(entry.key)
        if entry.is_array:
            path.append(str(entry.array_index))
    path.append(current_key)
    return path


def _skip_quoted_string(
    content: str,
    pos: int,
    quote: str,
    line: int,
) -> _PosLine:
    """Advance past a quoted string body. Returns position after the closing quote."""
    length = len(content)
    while pos < length:
        ch = content[pos]
        if ch == "\\":
            if pos + 1 < length and content[pos + 1] == "\n":
                line += 1
            pos += 2
            continue
        if ch == "\n":
            line += 1
            pos += 1
            continue
        if ch == quote:
            return _PosLine(pos=pos + 1, line=line)
        pos += 1
    return _PosLine(pos=pos, line=line)


def _is_ident_start(ch: str) -> bool:
    return ch.isalpha() or ch in {"_", "$"}


def _is_ident_char(ch: str) -> bool:
    return ch.isalnum() or ch in {"_", "$"}


def _skip_identifier(content: str, pos: int, length: int) -> int:
    while pos < length and _is_ident_char(content[pos]):
        pos += 1
    return pos


def _skip_line_comment(content: str, pos: int, length: int) -> int:
    while pos < length and content[pos] != "\n":
        pos += 1
    return pos


def _skip_block_comment(
    content: str,
    pos: int,
    length: int,
    line: int,
) -> _PosLine:
    while pos < length:
        if content[pos] == "\n":
            line += 1
        elif content[pos] == "*" and pos + 1 < length and content[pos + 1] == "/":
            return _PosLine(pos=pos + 2, line=line)
        pos += 1
    return _PosLine(pos=pos, line=line)


def _skip_ws_and_comments(
    content: str,
    pos: int,
    length: int,
    line: int,
) -> _PosLine:
    while pos < length:
        ch = content[pos]
        if ch in " \t\r":
            pos += 1
            continue
        if ch == "\n":
            line += 1
            pos += 1
            continue
        if ch == "/" and pos + 1 < length and content[pos + 1] == "/":
            pos = _skip_line_comment(content, pos + 2, length)
            continue
        if ch == "/" and pos + 1 < length and content[pos + 1] == "*":
            block = _skip_block_comment(content, pos + 2, length, line)
            pos = block.pos
            line = block.line
            continue
        return _PosLine(pos=pos, line=line)
    return _PosLine(pos=pos, line=line)


def _skip_scalar(content: str, pos: int, length: int) -> int:
    while pos < length and content[pos] not in " \t\r\n,}]":
        pos += 1
    return pos



import logging
from collections.abc import Callable
from dataclasses import asdict, fields, is_dataclass
from pathlib import Path
from typing import cast

from dature.deep_merge import deep_merge, deep_merge_last_wins, raise_on_conflict
from dature.error_formatter import ErrorContext, handle_load_errors, read_file_content
from dature.errors import DatureConfigError
from dature.load_report import (
    FieldOrigin,
    LoadReport,
    SourceEntry,
    attach_load_report,
    compute_field_origins,
    get_load_report,
)
from dature.metadata import FieldMergeStrategy, LoadMetadata, MergeMetadata, MergeStrategy
from dature.patcher import ensure_retort, merge_fields
from dature.predicate import build_field_merge_map
from dature.sources_loader.base import ILoader
from dature.sources_loader.resolver import get_loader_type, resolve_loader
from dature.types import JSONValue
from dature.validators.protocols import DataclassInstance

logger = logging.getLogger("dature")


def _build_error_ctx(metadata: LoadMetadata, dataclass_name: str) -> ErrorContext:
    loader_type = get_loader_type(metadata)
    error_file_path = Path(metadata.file_) if metadata.file_ else None
    return ErrorContext(
        dataclass_name=dataclass_name,
        loader_type=loader_type,
        file_path=error_file_path,
        prefix=metadata.prefix,
        split_symbols=metadata.split_symbols,
    )


def _resolve_loader_for_source(
    *,
    loaders: tuple[ILoader, ...] | None,
    index: int,
    source_meta: LoadMetadata,
) -> ILoader:
    if loaders is not None:
        return loaders[index]
    return resolve_loader(source_meta)


def _log_merge_step(
    *,
    dataclass_name: str,
    step_idx: int,
    strategy: MergeStrategy,
    before: JSONValue,
    source_data: JSONValue,
    after: JSONValue,
) -> None:
    if isinstance(before, dict) and isinstance(source_data, dict) and isinstance(after, dict):
        added_keys = set(source_data.keys()) - set(before.keys())
        overwritten_keys = set(source_data.keys()) & set(before.keys())
        logger.debug(
            "[%s] Merge step %d (strategy=%s): added=%s, overwritten=%s",
            dataclass_name,
            step_idx,
            strategy.value,
            sorted(added_keys),
            sorted(overwritten_keys),
        )
    logger.debug(
        "[%s] State after step %d: %s",
        dataclass_name,
        step_idx,
        after,
    )


def _log_field_origins(
    *,
    dataclass_name: str,
    field_origins: tuple[FieldOrigin, ...],
) -> None:
    for origin in field_origins:
        logger.debug(
            "[%s] Field '%s' = %r  <-- source %d (%s)",
            dataclass_name,
            origin.key,
            origin.value,
            origin.source_index,
            origin.source_file or "<env>",
        )


def _build_merge_report(
    *,
    dataclass_name: str,
    strategy: MergeStrategy,
    source_entries: tuple[SourceEntry, ...],
    field_origins: tuple[FieldOrigin, ...],
    merged_data: JSONValue,
) -> LoadReport:
    return LoadReport(
        dataclass_name=dataclass_name,
        strategy=strategy,
        sources=source_entries,
        field_origins=field_origins,
        merged_data=merged_data,
    )


def _merge_raw_dicts(
    *,
    raw_dicts: list[JSONValue],
    strategy: MergeStrategy,
    dataclass_name: str,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
) -> JSONValue:
    merged: JSONValue = {}
    for step_idx, raw in enumerate(raw_dicts):
        before = merged
        if strategy == MergeStrategy.RAISE_ON_CONFLICT:
            merged = deep_merge_last_wins(merged, raw, field_merge_map=field_merge_map)
        else:
            merged = deep_merge(merged, raw, strategy=strategy, field_merge_map=field_merge_map)

        _log_merge_step(
            dataclass_name=dataclass_name,
            step_idx=step_idx,
            strategy=strategy,
            before=before,
            source_data=raw,
            after=merged,
        )

    return merged


def _load_and_merge[T](
    *,
    merge_meta: MergeMetadata,
    dataclass_: type[T],
    loaders: tuple[ILoader, ...] | None = None,
    debug: bool = False,
) -> T:
    raw_dicts: list[JSONValue] = []
    source_ctxs: list[tuple[ErrorContext, str | None]] = []
    source_entries: list[SourceEntry] = []
    last_loader: ILoader | None = None

    for i, source_meta in enumerate(merge_meta.sources):
        loader_instance = _resolve_loader_for_source(loaders=loaders, index=i, source_meta=source_meta)
        file_path = Path(source_meta.file_) if source_meta.file_ else Path()
        error_ctx = _build_error_ctx(source_meta, dataclass_.__name__)

        def _load_raw(li: ILoader = loader_instance, fp: Path = file_path) -> JSONValue:
            return li.load_raw(fp)

        raw = handle_load_errors(
            func=_load_raw,
            ctx=error_ctx,
        )
        raw_dicts.append(raw)

        loader_type = get_loader_type(source_meta)

        logger.debug(
            "[%s] Source %d loaded: loader=%s, file=%s, keys=%s",
            dataclass_.__name__,
            i,
            loader_type,
            source_meta.file_ or "<env>",
            sorted(raw.keys()) if isinstance(raw, dict) else "<non-dict>",
        )
        logger.debug(
            "[%s] Source %d raw data: %s",
            dataclass_.__name__,
            i,
            raw,
        )

        source_entries.append(
            SourceEntry(
                index=i,
                file_path=source_meta.file_,
                loader_type=loader_type,
                raw_data=raw,
            ),
        )

        file_content = read_file_content(error_ctx.file_path)
        source_ctxs.append((error_ctx, file_content))
        last_loader = loader_instance

    if last_loader is None:
        msg = "MergeMetadata.sources must not be empty"
        raise ValueError(msg)

    field_merge_map: dict[str, FieldMergeStrategy] | None = None
    if merge_meta.field_merges:
        field_merge_map = build_field_merge_map(merge_meta.field_merges)

    if merge_meta.strategy == MergeStrategy.RAISE_ON_CONFLICT:
        raise_on_conflict(raw_dicts, source_ctxs, dataclass_.__name__, field_merge_map=field_merge_map)

    merged = _merge_raw_dicts(
        raw_dicts=raw_dicts,
        strategy=merge_meta.strategy,
        dataclass_name=dataclass_.__name__,
        field_merge_map=field_merge_map,
    )

    logger.debug(
        "[%s] Merged result (strategy=%s, %d sources): %s",
        dataclass_.__name__,
        merge_meta.strategy.value,
        len(raw_dicts),
        merged,
    )

    frozen_entries = tuple(source_entries)
    field_origins = compute_field_origins(
        raw_dicts=raw_dicts,
        source_entries=frozen_entries,
        strategy=merge_meta.strategy,
    )

    _log_field_origins(
        dataclass_name=dataclass_.__name__,
        field_origins=field_origins,
    )

    report: LoadReport | None = None
    if debug:
        report = _build_merge_report(
            dataclass_name=dataclass_.__name__,
            strategy=merge_meta.strategy,
            source_entries=frozen_entries,
            field_origins=field_origins,
            merged_data=merged,
        )

    last_error_ctx = source_ctxs[-1][0]
    try:
        result = handle_load_errors(
            func=lambda: last_loader.transform_to_dataclass(merged, dataclass_),
            ctx=last_error_ctx,
        )
    except DatureConfigError:
        if report is not None:
            attach_load_report(dataclass_, report)
        raise

    if report is not None:
        attach_load_report(result, report)

    return result


def merge_load_as_function[T](
    merge_meta: MergeMetadata,
    dataclass_: type[T],
    *,
    debug: bool = False,
) -> T:
    result = _load_and_merge(
        merge_meta=merge_meta,
        dataclass_=dataclass_,
        debug=debug,
    )

    last_meta = merge_meta.sources[-1]
    last_loader = resolve_loader(last_meta)
    validating_retort = last_loader.create_validating_retort(dataclass_)
    validation_loader = validating_retort.get_loader(dataclass_)
    result_dict = asdict(cast("DataclassInstance", result))

    last_error_ctx = _build_error_ctx(last_meta, dataclass_.__name__)
    try:
        handle_load_errors(
            func=lambda: validation_loader(result_dict),
            ctx=last_error_ctx,
        )
    except DatureConfigError:
        if debug:
            report = get_load_report(result)
            if report is not None:
                attach_load_report(dataclass_, report)
        raise

    return result


class _MergePatchContext:
    def __init__(
        self,
        *,
        merge_meta: MergeMetadata,
        cls: type[DataclassInstance],
        cache: bool,
        debug: bool,
    ) -> None:
        self.loaders = self._prepare_loaders(merge_meta=merge_meta, cls=cls)

        self.merge_meta = merge_meta
        self.cls = cls
        self.cache = cache
        self.debug = debug
        self.cached_data: DataclassInstance | None = None
        self.field_list = fields(cls)
        self.original_init = cls.__init__
        self.original_post_init = getattr(cls, "__post_init__", None)
        self.loading = False
        self.validating = False

        last_loader = self.loaders[-1]
        validating_retort = last_loader.create_validating_retort(cls)
        self.validation_loader = validating_retort.get_loader(cls)

        last_meta = merge_meta.sources[-1]
        self.error_ctx = _build_error_ctx(last_meta, cls.__name__)

    @staticmethod
    def _prepare_loaders(
        *,
        merge_meta: MergeMetadata,
        cls: type,
    ) -> tuple[ILoader, ...]:
        loaders: list[ILoader] = []
        for source_meta in merge_meta.sources:
            loader_instance = resolve_loader(source_meta)
            ensure_retort(loader_instance, cls)
            loaders.append(loader_instance)
        return tuple(loaders)


def _make_merge_new_init(ctx: _MergePatchContext) -> Callable[..., None]:
    def new_init(self: DataclassInstance, *args: object, **kwargs: object) -> None:
        if ctx.loading:
            ctx.original_init(self, *args, **kwargs)
            return

        if ctx.cache and ctx.cached_data is not None:
            loaded_data = ctx.cached_data
        else:
            ctx.loading = True
            try:
                loaded_data = _load_and_merge(
                    merge_meta=ctx.merge_meta,
                    dataclass_=ctx.cls,
                    loaders=ctx.loaders,
                    debug=ctx.debug,
                )
            finally:
                ctx.loading = False
            if ctx.cache:
                ctx.cached_data = loaded_data

        complete_kwargs = merge_fields(loaded_data, ctx.field_list, args, kwargs)
        ctx.original_init(self, *args, **complete_kwargs)

        if ctx.debug:
            report = get_load_report(loaded_data)
            if report is not None:
                attach_load_report(self, report)

        if ctx.original_post_init is None:
            self.__post_init__()  # type: ignore[attr-defined]

    return new_init


def _make_merge_new_post_init(ctx: _MergePatchContext) -> Callable[..., None]:
    def new_post_init(self: DataclassInstance) -> None:
        if ctx.loading:
            return

        if ctx.validating:
            return

        if ctx.original_post_init is not None:
            ctx.original_post_init(self)

        ctx.validating = True
        try:
            obj_dict = asdict(self)
            handle_load_errors(
                func=lambda: ctx.validation_loader(obj_dict),
                ctx=ctx.error_ctx,
            )
        finally:
            ctx.validating = False

    return new_post_init


def merge_make_decorator(
    merge_meta: MergeMetadata,
    *,
    cache: bool = True,
    debug: bool = False,
) -> Callable[[type[DataclassInstance]], type[DataclassInstance]]:
    def decorator(cls: type[DataclassInstance]) -> type[DataclassInstance]:
        if not is_dataclass(cls):
            msg = f"{cls.__name__} must be a dataclass"
            raise TypeError(msg)

        ctx = _MergePatchContext(
            merge_meta=merge_meta,
            cls=cls,
            cache=cache,
            debug=debug,
        )
        cls.__init__ = _make_merge_new_init(ctx)  # type: ignore[method-assign]
        cls.__post_init__ = _make_merge_new_post_init(ctx)  # type: ignore[attr-defined]
        return cls

    return decorator



import logging
from collections.abc import Callable
from dataclasses import asdict, fields, is_dataclass
from pathlib import Path
from typing import Any

from dature.error_formatter import ErrorContext, handle_load_errors
from dature.errors import DatureConfigError
from dature.load_report import (
    FieldOrigin,
    LoadReport,
    SourceEntry,
    attach_load_report,
)
from dature.metadata import LoadMetadata
from dature.sources_loader.base import ILoader
from dature.sources_loader.resolver import get_loader_type
from dature.types import JSONValue
from dature.validators.protocols import DataclassInstance

logger = logging.getLogger("dature")


def merge_fields(
    loaded_data: Any,  # noqa: ANN401
    field_list: tuple[Any, ...],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> dict[str, Any]:
    explicit_fields = set(kwargs.keys())
    for i, _ in enumerate(args):
        if i < len(field_list):
            explicit_fields.add(field_list[i].name)

    complete_kwargs = dict(kwargs)
    for field in field_list:
        if field.name not in explicit_fields:
            complete_kwargs[field.name] = getattr(loaded_data, field.name)

    return complete_kwargs


def ensure_retort(loader_instance: ILoader, cls: type) -> None:
    """Creates a replacement response to __init__ so that Adaptix sees the original signature."""
    if cls not in loader_instance.retorts:
        loader_instance.retorts[cls] = loader_instance.create_retort()
    loader_instance.retorts[cls].get_loader(cls)


def _log_single_source_load(
    *,
    dataclass_name: str,
    loader_type: str,
    file_path: str,
    data: JSONValue,
) -> None:
    logger.debug(
        "[%s] Single-source load: loader=%s, file=%s",
        dataclass_name,
        loader_type,
        file_path,
    )
    logger.debug(
        "[%s] Loaded data: %s",
        dataclass_name,
        data,
    )


def _build_single_source_report(
    *,
    dataclass_name: str,
    loader_type: str,
    file_path: str | None,
    raw_data: JSONValue,
) -> LoadReport:
    source = SourceEntry(
        index=0,
        file_path=file_path,
        loader_type=loader_type,
        raw_data=raw_data,
    )

    origins: list[FieldOrigin] = []
    if isinstance(raw_data, dict):
        for key, value in sorted(raw_data.items()):
            origins.append(
                FieldOrigin(
                    key=key,
                    value=value,
                    source_index=0,
                    source_file=file_path,
                    source_loader_type=loader_type,
                ),
            )

    return LoadReport(
        dataclass_name=dataclass_name,
        strategy=None,
        sources=(source,),
        field_origins=tuple(origins),
        merged_data=raw_data,
    )


class _PatchContext:
    def __init__(
        self,
        *,
        loader_instance: ILoader,
        file_path: Path,
        cls: type[DataclassInstance],
        metadata: LoadMetadata,
        cache: bool,
        debug: bool,
    ) -> None:
        ensure_retort(loader_instance, cls)
        validating_retort = loader_instance.create_validating_retort(cls)

        self.loader_instance = loader_instance
        self.file_path = file_path
        self.cls = cls
        self.metadata = metadata
        self.cache = cache
        self.debug = debug
        self.cached_data: DataclassInstance | None = None
        self.field_list = fields(cls)
        self.original_init = cls.__init__
        self.original_post_init = getattr(cls, "__post_init__", None)
        self.validation_loader = validating_retort.get_loader(cls)
        self.validating = False
        self.loading = False

        self.loader_type = get_loader_type(metadata)
        self.error_file_path = Path(metadata.file_) if metadata.file_ else None
        self.prefix = metadata.prefix
        self.split_symbols = metadata.split_symbols
        self.error_ctx = ErrorContext(
            dataclass_name=cls.__name__,
            loader_type=self.loader_type,
            file_path=self.error_file_path,
            prefix=self.prefix,
            split_symbols=self.split_symbols,
        )


def _make_new_init(ctx: _PatchContext) -> Callable[..., None]:
    def new_init(self: DataclassInstance, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        if ctx.loading:
            ctx.original_init(self, *args, **kwargs)
            return

        if ctx.cache and ctx.cached_data is not None:
            loaded_data = ctx.cached_data
        else:
            ctx.loading = True
            try:
                loaded_data = handle_load_errors(
                    func=lambda: ctx.loader_instance.load(ctx.file_path, ctx.cls),
                    ctx=ctx.error_ctx,
                )
            finally:
                ctx.loading = False

            _log_single_source_load(
                dataclass_name=ctx.cls.__name__,
                loader_type=ctx.loader_type,
                file_path=str(ctx.file_path),
                data=asdict(loaded_data),
            )

            if ctx.cache:
                ctx.cached_data = loaded_data

        complete_kwargs = merge_fields(loaded_data, ctx.field_list, args, kwargs)
        ctx.original_init(self, *args, **complete_kwargs)

        if ctx.debug:
            result_dict = asdict(self)
            report = _build_single_source_report(
                dataclass_name=ctx.cls.__name__,
                loader_type=ctx.loader_type,
                file_path=str(ctx.file_path) if ctx.metadata.file_ is not None else None,
                raw_data=result_dict,
            )
            attach_load_report(self, report)

        if ctx.original_post_init is None:
            self.__post_init__()  # type: ignore[attr-defined]

    return new_init


def _make_new_post_init(ctx: _PatchContext) -> Callable[..., None]:
    def new_post_init(self: DataclassInstance) -> None:
        if ctx.loading:
            return

        if ctx.validating:
            return

        if ctx.original_post_init is not None:
            ctx.original_post_init(self)

        ctx.validating = True
        try:
            obj_dict = asdict(self)
            handle_load_errors(
                func=lambda: ctx.validation_loader(obj_dict),
                ctx=ctx.error_ctx,
            )
        finally:
            ctx.validating = False

    return new_post_init


def load_as_function(
    *,
    loader_instance: ILoader,
    file_path: Path,
    dataclass_: type[DataclassInstance],
    metadata: LoadMetadata,
    debug: bool,
) -> DataclassInstance:
    loader_type = get_loader_type(metadata)
    error_file_path = Path(metadata.file_) if metadata.file_ else None
    error_ctx = ErrorContext(
        dataclass_name=dataclass_.__name__,
        loader_type=loader_type,
        file_path=error_file_path,
        prefix=metadata.prefix,
        split_symbols=metadata.split_symbols,
    )

    raw_data = handle_load_errors(
        func=lambda: loader_instance.load_raw(file_path),
        ctx=error_ctx,
    )

    report: LoadReport | None = None
    if debug:
        report = _build_single_source_report(
            dataclass_name=dataclass_.__name__,
            loader_type=loader_type,
            file_path=metadata.file_,
            raw_data=raw_data,
        )

    _log_single_source_load(
        dataclass_name=dataclass_.__name__,
        loader_type=loader_type,
        file_path=str(file_path),
        data=raw_data if isinstance(raw_data, dict) else {},
    )

    try:
        result = handle_load_errors(
            func=lambda: loader_instance.transform_to_dataclass(raw_data, dataclass_),
            ctx=error_ctx,
        )
    except DatureConfigError:
        if report is not None:
            attach_load_report(dataclass_, report)
        raise

    result_dict = asdict(result)

    validating_retort = loader_instance.create_validating_retort(dataclass_)
    validation_loader = validating_retort.get_loader(dataclass_)

    try:
        handle_load_errors(
            func=lambda: validation_loader(result_dict),
            ctx=error_ctx,
        )
    except DatureConfigError:
        if report is not None:
            attach_load_report(dataclass_, report)
        raise

    if report is not None:
        attach_load_report(result, report)

    return result


def make_decorator(
    *,
    loader_instance: ILoader,
    file_path: Path,
    metadata: LoadMetadata,
    cache: bool = True,
    debug: bool = False,
) -> Callable[[type[DataclassInstance]], type[DataclassInstance]]:
    def decorator(cls: type[DataclassInstance]) -> type[DataclassInstance]:
        if not is_dataclass(cls):
            msg = f"{cls.__name__} must be a dataclass"
            raise TypeError(msg)

        ctx = _PatchContext(
            loader_instance=loader_instance,
            file_path=file_path,
            cls=cls,
            metadata=metadata,
            cache=cache,
            debug=debug,
        )
        cls.__init__ = _make_new_init(ctx)  # type: ignore[method-assign]
        cls.__post_init__ = _make_new_post_init(ctx)  # type: ignore[attr-defined]
        return cls

    return decorator



from dataclasses import dataclass
from enum import StrEnum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dature.sources_loader.resolver import LoaderType
    from dature.types import DotSeparatedPath, FieldMapping, NameStyle
    from dature.validators.protocols import ValidatorProtocol


class MergeStrategy(StrEnum):
    LAST_WINS = "last_wins"
    FIRST_WINS = "first_wins"
    RAISE_ON_CONFLICT = "raise_on_conflict"


class FieldMergeStrategy(StrEnum):
    FIRST_WINS = "first_wins"
    LAST_WINS = "last_wins"
    APPEND = "append"
    APPEND_UNIQUE = "append_unique"
    PREPEND = "prepend"
    PREPEND_UNIQUE = "prepend_unique"
    MAX = "max"
    MIN = "min"


@dataclass(frozen=True, slots=True)
class MergeRule:
    predicate: object
    strategy: FieldMergeStrategy


@dataclass(frozen=True, slots=True, kw_only=True)
class LoadMetadata:
    file_: str | None = None
    loader: "LoaderType | None" = None
    prefix: "DotSeparatedPath | None" = None
    split_symbols: str = "__"
    name_style: "NameStyle | None" = None
    field_mapping: "FieldMapping | None" = None
    root_validators: "tuple[ValidatorProtocol, ...] | None" = None
    enable_expand_env_vars: bool = True


@dataclass(frozen=True, slots=True, kw_only=True)
class MergeMetadata:
    sources: tuple[LoadMetadata, ...]
    strategy: MergeStrategy = MergeStrategy.LAST_WINS
    field_merges: tuple[MergeRule, ...] = ()



from dature.error_formatter import ErrorContext, resolve_source_location
from dature.errors import FieldErrorInfo, MergeConflictError, SourceLocation
from dature.metadata import FieldMergeStrategy, MergeStrategy
from dature.types import JSONValue

_MIN_CONFLICT_SOURCES = 2


def _deduplicate_list(items: list[JSONValue]) -> list[JSONValue]:
    seen: list[JSONValue] = []
    for item in items:
        found = False
        for s in seen:
            if s == item:
                found = True
                break
        if not found:
            seen.append(item)
    return seen


def _ensure_both_lists(
    base: JSONValue,
    override: JSONValue,
    strategy_name: str,
) -> tuple[list[JSONValue], list[JSONValue]]:
    if not isinstance(base, list) or not isinstance(override, list):
        base_type = type(base).__name__
        override_type = type(override).__name__
        msg = f"{strategy_name} strategy requires both values to be lists, got {base_type} and {override_type}"
        raise TypeError(msg)
    return base, override


def _apply_list_merge(
    base: JSONValue,
    override: JSONValue,
    strategy: FieldMergeStrategy,
) -> list[JSONValue]:
    if strategy == FieldMergeStrategy.APPEND:
        b, o = _ensure_both_lists(base, override, "APPEND")
        return list(b) + list(o)

    if strategy == FieldMergeStrategy.APPEND_UNIQUE:
        b, o = _ensure_both_lists(base, override, "APPEND_UNIQUE")
        return _deduplicate_list(list(b) + list(o))

    if strategy == FieldMergeStrategy.PREPEND:
        b, o = _ensure_both_lists(base, override, "PREPEND")
        return list(o) + list(b)

    # PREPEND_UNIQUE
    b, o = _ensure_both_lists(base, override, "PREPEND_UNIQUE")
    return _deduplicate_list(list(o) + list(b))


_COMPARABLE = (int, float, str)


def apply_field_merge(
    base: JSONValue,
    override: JSONValue,
    strategy: FieldMergeStrategy,
) -> JSONValue:
    if strategy == FieldMergeStrategy.FIRST_WINS:
        return base

    if strategy == FieldMergeStrategy.LAST_WINS:
        return override

    if strategy in (
        FieldMergeStrategy.APPEND,
        FieldMergeStrategy.APPEND_UNIQUE,
        FieldMergeStrategy.PREPEND,
        FieldMergeStrategy.PREPEND_UNIQUE,
    ):
        return _apply_list_merge(base, override, strategy)

    if strategy == FieldMergeStrategy.MAX:
        if isinstance(base, _COMPARABLE) and isinstance(override, _COMPARABLE):
            return max(base, override)
        msg = f"MAX strategy requires comparable values, got {type(base).__name__} and {type(override).__name__}"
        raise TypeError(msg)

    # MIN
    if isinstance(base, _COMPARABLE) and isinstance(override, _COMPARABLE):
        return min(base, override)
    msg = f"MIN strategy requires comparable values, got {type(base).__name__} and {type(override).__name__}"
    raise TypeError(msg)


def deep_merge_last_wins(
    base: JSONValue,
    override: JSONValue,
    *,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
    _path: str = "",
) -> JSONValue:
    if field_merge_map is not None and _path in field_merge_map:
        return apply_field_merge(base, override, field_merge_map[_path])

    if isinstance(base, dict) and isinstance(override, dict):
        result = dict(base)
        for key, value in override.items():
            child_path = f"{_path}.{key}" if _path else key
            if key in result:
                result[key] = deep_merge_last_wins(
                    result[key],
                    value,
                    field_merge_map=field_merge_map,
                    _path=child_path,
                )
            else:
                result[key] = value
        return result
    return override


def deep_merge_first_wins(
    base: JSONValue,
    override: JSONValue,
    *,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
    _path: str = "",
) -> JSONValue:
    if field_merge_map is not None and _path in field_merge_map:
        return apply_field_merge(base, override, field_merge_map[_path])

    if isinstance(base, dict) and isinstance(override, dict):
        result = dict(base)
        for key, value in override.items():
            child_path = f"{_path}.{key}" if _path else key
            if key in result:
                result[key] = deep_merge_first_wins(
                    result[key],
                    value,
                    field_merge_map=field_merge_map,
                    _path=child_path,
                )
            else:
                result[key] = value
        return result
    return base


def _collect_conflicts(
    dicts: list[JSONValue],
    source_contexts: list[tuple[ErrorContext, str | None]],
    path: list[str],
    conflicts: list[tuple[list[str], list[tuple[int, JSONValue]]]],
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
) -> None:
    key_sources: dict[str, list[tuple[int, JSONValue]]] = {}

    for i, d in enumerate(dicts):
        if not isinstance(d, dict):
            continue
        for key, value in d.items():
            if key not in key_sources:
                key_sources[key] = []
            key_sources[key].append((i, value))

    for key, sources in key_sources.items():
        if len(sources) < _MIN_CONFLICT_SOURCES:
            continue

        field_path = ".".join([*path, key])
        if field_merge_map is not None and field_path in field_merge_map:
            continue

        values = [v for _, v in sources]

        nested_dicts = [v for v in values if isinstance(v, dict)]
        if len(nested_dicts) == len(sources):
            _collect_conflicts(
                values,
                [source_contexts[i] for i, _ in sources],
                [*path, key],
                conflicts,
                field_merge_map=field_merge_map,
            )
            continue

        if all(v == values[0] for v in values[1:]):
            continue

        conflicts.append(([*path, key], sources))


def raise_on_conflict(
    dicts: list[JSONValue],
    source_ctxs: list[tuple[ErrorContext, str | None]],
    dataclass_name: str,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
) -> None:
    conflicts: list[tuple[list[str], list[tuple[int, JSONValue]]]] = []
    _collect_conflicts(dicts, source_ctxs, [], conflicts, field_merge_map=field_merge_map)

    if not conflicts:
        return

    conflict_errors: list[tuple[FieldErrorInfo, list[SourceLocation]]] = []
    for field_path, sources in conflicts:
        info = FieldErrorInfo(
            field_path=field_path,
            message="Conflicting values in multiple sources",
            input_value=None,
        )
        locations: list[SourceLocation] = []
        for source_idx, _ in sources:
            ctx, file_content = source_ctxs[source_idx]
            loc = resolve_source_location(field_path, ctx, file_content)
            locations.append(loc)
        conflict_errors.append((info, locations))

    raise MergeConflictError(conflict_errors, dataclass_name)


def deep_merge(
    base: JSONValue,
    override: JSONValue,
    *,
    strategy: MergeStrategy,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
) -> JSONValue:
    if strategy == MergeStrategy.LAST_WINS:
        return deep_merge_last_wins(base, override, field_merge_map=field_merge_map)
    if strategy == MergeStrategy.FIRST_WINS:
        return deep_merge_first_wins(base, override, field_merge_map=field_merge_map)
    msg = "Use merge_sources for RAISE_ON_CONFLICT strategy"
    raise ValueError(msg)



from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True)
class FieldErrorInfo:
    field_path: list[str]
    message: str
    input_value: object | None


@dataclass(frozen=True)
class SourceLocation:
    source_type: str
    file_path: Path | None
    line_number: int | None
    line_content: str | None
    env_var_name: str | None


@dataclass(frozen=True)
class FieldError:
    error: FieldErrorInfo
    location: SourceLocation | None


def _format_location(loc: SourceLocation) -> list[str]:
    lines: list[str] = []

    if loc.source_type == "env":
        if loc.env_var_name is not None:
            lines.append(f"   └── ENV '{loc.env_var_name}'")
        return lines

    if loc.source_type == "envfile":
        location_str = f"   └── ENV FILE '{loc.file_path}'"
        if loc.env_var_name is not None:
            location_str += f", var '{loc.env_var_name}'"
        lines.append(location_str)
        if loc.line_content is not None:
            lines.append(f"       {loc.line_content}")
        return lines

    location_str = f"   └── FILE '{loc.file_path}'"
    if loc.line_number is not None:
        location_str += f", line {loc.line_number}"
    lines.append(location_str)
    if loc.line_content is not None:
        lines.append(f"       {loc.line_content}")

    return lines


class DatureConfigError(Exception):
    def __init__(self, errors: list[FieldError], dataclass_name: str) -> None:
        self.errors = errors
        self.dataclass_name = dataclass_name
        super().__init__(self._format_message())

    def _format_message(self) -> str:
        lines: list[str] = []
        lines.append(f"{self.dataclass_name} loading errors ({len(self.errors)})")
        lines.append("")

        for field_error in self.errors:
            path_str = ".".join(field_error.error.field_path)
            if not path_str:
                path_str = "<root>"
            lines.append(f"  [{path_str}]  {field_error.error.message}")

            if field_error.location is not None:
                lines.extend(_format_location(field_error.location))

            lines.append("")

        return "\n".join(lines)


class MergeConflictError(DatureConfigError):
    def __init__(
        self,
        conflicts: list[tuple[FieldErrorInfo, list[SourceLocation]]],
        dataclass_name: str,
    ) -> None:
        self.conflicts = conflicts
        errors = [FieldError(error=info, location=locations[0] if locations else None) for info, locations in conflicts]
        super().__init__(errors, dataclass_name)

    def _format_message(self) -> str:
        lines: list[str] = []
        lines.append(f"{self.dataclass_name} merge conflicts ({len(self.conflicts)})")
        lines.append("")

        for info, locations in self.conflicts:
            path_str = ".".join(info.field_path)
            if not path_str:
                path_str = "<root>"
            lines.append(f"  [{path_str}]  {info.message}")

            for loc in locations:
                lines.extend(_format_location(loc))

            lines.append("")

        return "\n".join(lines)



class YamlPathFinder:
    def __init__(self, content: str):
        self.lines = content.splitlines()

    def find_line(self, target_path: list[str]) -> int:
        stack: list[dict[str, str | int]] = []
        for i, line in enumerate(self.lines, 1):
            stripped = line.lstrip()
            if not stripped or stripped.startswith(("#", "-")):
                continue

            indent = len(line) - len(stripped)
            # Separate by the first colon
            key_part, sep, _ = stripped.partition(":")
            if not sep:
                continue

            key = key_part.strip().strip("\"'")

            while stack and int(stack[-1]["indent"]) >= indent:
                stack.pop()

            stack.append({"key": key, "indent": indent})
            if [item["key"] for item in stack] == target_path:
                return i
        return -1



class TomlPathFinder:
    def __init__(self, content: str) -> None:
        self.lines = content.splitlines()

    def find_line(self, target_path: list[str]) -> int:
        target_key = target_path[-1]
        target_parents = target_path[:-1]
        current_section: list[str] = []
        in_multiline: str | None = None

        for i, line in enumerate(self.lines, 1):
            if in_multiline is not None:
                if in_multiline in line:
                    in_multiline = None
                continue

            stripped = line.strip()
            if not stripped or stripped.startswith(("#", ";")):
                continue

            # Working with sections [section] or [parent.child]
            if stripped.startswith("[") and "]" in stripped:
                section_name = stripped.strip("[] ")
                current_section = [p.strip() for p in section_name.split(".")]
                continue

            # Working with keys key = value
            if "=" not in stripped:
                continue

            key, _, value = stripped.partition("=")
            key = key.strip().strip("\"'")
            value = value.strip()

            in_multiline = _detect_unclosed_multiline(value)
            if in_multiline is not None:
                continue

            if key == target_key and current_section == target_parents:
                return i
        return -1


def _detect_unclosed_multiline(value: str) -> str | None:
    for delimiter in ('"""', "'''"):
        if delimiter not in value:
            continue
        # odd number of delimiters means unclosed
        count = value.count(delimiter)
        if count % 2 == 1:
            return delimiter
    return None



from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path

from adaptix.load_error import (
    AggregateLoadError,
    BadVariantLoadError,
    ExtraFieldsLoadError,
    LoadError,
    LoadExceptionGroup,
    NoRequiredFieldsLoadError,
    TypeLoadError,
    ValidationLoadError,
    ValueLoadError,
)
from adaptix.struct_trail import get_trail

from dature.errors import DatureConfigError, FieldError, FieldErrorInfo, SourceLocation
from dature.source_locators.ini_ import TablePathFinder
from dature.source_locators.json5_ import Json5PathFinder
from dature.source_locators.json_ import JsonPathFinder
from dature.source_locators.toml_ import TomlPathFinder
from dature.source_locators.yaml_ import YamlPathFinder

_PATH_FINDER_MAP: dict[
    str,
    type[YamlPathFinder | JsonPathFinder | Json5PathFinder | TomlPathFinder | TablePathFinder],
] = {
    "yaml": YamlPathFinder,
    "yaml1.1": YamlPathFinder,
    "yaml1.2": YamlPathFinder,
    "json": JsonPathFinder,
    "json5": Json5PathFinder,
    "toml": TomlPathFinder,
    "ini": TablePathFinder,
}


@dataclass(frozen=True)
class ErrorContext:
    dataclass_name: str
    loader_type: str
    file_path: Path | None
    prefix: str | None
    split_symbols: str


def _describe_error(exc: BaseException) -> str:
    if isinstance(exc, (ValidationLoadError, ValueLoadError)):
        return str(exc.msg)

    if isinstance(exc, TypeLoadError):
        return f"Expected {exc.expected_type.__name__}, got {type(exc.input_value).__name__}"

    if isinstance(exc, ExtraFieldsLoadError):
        field_names = ", ".join(sorted(exc.fields))
        return f"Unknown field(s): {field_names}"

    if isinstance(exc, BadVariantLoadError):
        return f"Invalid variant: {exc.input_value!r}"

    return str(exc)


def _walk_exception(
    exc: BaseException,
    parent_path: list[str],
    result: list[FieldErrorInfo],
) -> None:
    trail = list(get_trail(exc))
    current_path = parent_path + [str(elem) for elem in trail]

    if isinstance(exc, LoadExceptionGroup):
        for sub_exc in exc.exceptions:
            _walk_exception(sub_exc, current_path, result)
        return

    if isinstance(exc, NoRequiredFieldsLoadError):
        result.extend(
            FieldErrorInfo(
                field_path=[*current_path, field_name],
                message="Missing required field",
                input_value=None,
            )
            for field_name in sorted(exc.fields)
        )
        return

    result.append(
        FieldErrorInfo(
            field_path=current_path,
            message=_describe_error(exc),
            input_value=getattr(exc, "input_value", None),
        ),
    )


def extract_field_errors(exc: BaseException) -> list[FieldErrorInfo]:
    result: list[FieldErrorInfo] = []
    _walk_exception(exc, [], result)
    return result


def read_file_content(file_path: Path | None) -> str | None:
    if file_path is None:
        return None

    with suppress(OSError):
        return file_path.read_text()

    return None


def _build_env_var_name(
    field_path: list[str],
    prefix: str | None,
    split_symbols: str,
) -> str:
    var_name = split_symbols.join(part.upper() for part in field_path)
    if prefix is not None:
        return prefix + var_name
    return var_name


def _build_search_path(field_path: list[str], prefix: str | None) -> list[str]:
    if not prefix:
        return field_path
    prefix_parts = prefix.split(".")
    return prefix_parts + field_path


def _find_env_line(content: str, var_name: str) -> SourceLocation:
    for i, line in enumerate(content.splitlines(), 1):
        stripped = line.strip()
        if not stripped or stripped.startswith("#"):
            continue
        if "=" not in stripped:
            continue
        key = stripped.split("=", 1)[0].strip()
        if key == var_name:
            return SourceLocation(
                source_type="envfile",
                file_path=None,
                line_number=i,
                line_content=stripped,
                env_var_name=var_name,
            )
    return SourceLocation(
        source_type="envfile",
        file_path=None,
        line_number=None,
        line_content=None,
        env_var_name=var_name,
    )


def _empty_file_location(loader_type: str, file_path: Path | None) -> SourceLocation:
    return SourceLocation(
        source_type=loader_type,
        file_path=file_path,
        line_number=None,
        line_content=None,
        env_var_name=None,
    )


def _resolve_file_location(
    field_path: list[str],
    loader_type: str,
    file_path: Path | None,
    file_content: str | None,
    prefix: str | None,
) -> SourceLocation:
    if file_content is None or not field_path:
        return _empty_file_location(loader_type, file_path)

    finder_class = _PATH_FINDER_MAP.get(loader_type)
    if finder_class is None:
        return _empty_file_location(loader_type, file_path)

    search_path = _build_search_path(field_path, prefix)
    found_line = finder_class(file_content).find_line(search_path)
    if found_line == -1:
        return _empty_file_location(loader_type, file_path)

    lines = file_content.splitlines()
    line_content = None
    if 0 < found_line <= len(lines):
        line_content = lines[found_line - 1].strip()

    return SourceLocation(
        source_type=loader_type,
        file_path=file_path,
        line_number=found_line,
        line_content=line_content,
        env_var_name=None,
    )


def resolve_source_location(
    field_path: list[str],
    ctx: ErrorContext,
    file_content: str | None,
) -> SourceLocation:
    if ctx.loader_type == "env":
        env_var_name = _build_env_var_name(field_path, ctx.prefix, ctx.split_symbols)
        return SourceLocation(
            source_type="env",
            file_path=None,
            line_number=None,
            line_content=None,
            env_var_name=env_var_name,
        )

    if ctx.loader_type == "envfile":
        env_var_name = _build_env_var_name(field_path, ctx.prefix, ctx.split_symbols)
        if file_content is not None:
            location = _find_env_line(file_content, env_var_name)
            return SourceLocation(
                source_type="envfile",
                file_path=ctx.file_path,
                line_number=location.line_number,
                line_content=location.line_content,
                env_var_name=env_var_name,
            )
        return SourceLocation(
            source_type="envfile",
            file_path=ctx.file_path,
            line_number=None,
            line_content=None,
            env_var_name=env_var_name,
        )

    return _resolve_file_location(field_path, ctx.loader_type, ctx.file_path, file_content, ctx.prefix)


def handle_load_errors[T](
    *,
    func: Callable[[], T],
    ctx: ErrorContext,
) -> T:
    try:
        return func()
    except (AggregateLoadError, LoadError) as exc:
        file_content = read_file_content(ctx.file_path)
        field_errors = extract_field_errors(exc)
        enriched: list[FieldError] = []
        for fe in field_errors:
            location = resolve_source_location(fe.field_path, ctx, file_content)
            enriched.append(FieldError(error=fe, location=location))
        raise DatureConfigError(enriched, ctx.dataclass_name) from exc



import json
from datetime import date, datetime, time
from typing import Any, cast

# Expected number of time parts in HH:MM:SS format
TIME_PARTS_WITH_SECONDS = 3

# Expected number of time parts in HH:MM format
TIME_PARTS_WITHOUT_SECONDS = 2


def date_from_string(value: str) -> date:
    return date.fromisoformat(value)


def datetime_from_string(value: str) -> datetime:
    return datetime.fromisoformat(value)


def time_from_string(value: str) -> time:
    parts = value.split(":")
    if len(parts) == TIME_PARTS_WITH_SECONDS:
        return time(int(parts[0]), int(parts[1]), int(parts[2]))
    if len(parts) == TIME_PARTS_WITHOUT_SECONDS:
        return time(int(parts[0]), int(parts[1]))
    msg = f"Invalid time format: {value}"
    raise ValueError(msg)


def date_passthrough(value: date) -> date:
    return value


def datetime_passthrough(value: datetime) -> datetime:
    return value


def bytearray_from_string(value: str) -> bytearray:
    return bytearray(value, "utf-8")


def none_from_empty_string(value: str) -> None:
    if value == "":
        return
    msg = f"Cannot convert {value!r} to None"
    raise TypeError(msg)


def optional_from_empty_string(value: str) -> str | None:
    if value == "":
        return None
    return value


def bool_from_string(value: str) -> bool:
    lower = value.lower().strip()
    if lower in ("true", "1", "yes", "on"):
        return True
    if lower in ("false", "0", "no", "off", ""):
        return False
    msg = f"Cannot convert {value!r} to bool"
    raise TypeError(msg)


def bytearray_from_json_string(value: str) -> bytearray:
    if value == "":
        return bytearray()

    if value.startswith("["):
        items = json.loads(value)
        if not isinstance(items, list):
            msg = f"Expected list in JSON, got {type(items)}"
            raise TypeError(msg)
        return bytearray(items)

    return bytearray(value.encode("utf-8"))


def list_from_json_string(value: str) -> list[Any]:
    if value == "":
        return []

    return cast("list[Any]", json.loads(value))


def tuple_from_json_string(value: str) -> tuple[Any, ...]:
    if value == "":
        return ()

    parsed = json.loads(value)
    if not isinstance(parsed, list):
        msg = f"Expected list in JSON, got {type(parsed)}"
        raise TypeError(msg)
    result = [tuple(item) if isinstance(item, list) else item for item in parsed]
    return tuple(result)


def set_from_json_string(value: str) -> set[Any]:
    if value == "":
        return set()

    parsed = json.loads(value)
    if not isinstance(parsed, list):
        msg = f"Expected list in JSON, got {type(parsed)}"
        raise TypeError(msg)
    return set(parsed)


def frozenset_from_json_string(value: str) -> frozenset[Any]:
    if value == "":
        return frozenset()

    parsed = json.loads(value)
    if not isinstance(parsed, list):
        msg = f"Expected list in JSON, got {type(parsed)}"
        raise TypeError(msg)
    return frozenset(parsed)



from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent

import pytest
from adaptix import Retort
from adaptix.load_error import AggregateLoadError, LoadError

from dature import LoadMetadata, load
from dature.error_formatter import ErrorContext, extract_field_errors, resolve_source_location
from dature.errors import DatureConfigError, FieldError, FieldErrorInfo, SourceLocation


class TestExtractFieldErrors:
    def test_type_error(self):
        @dataclass
        class Config:
            timeout: int

        r = Retort(strict_coercion=False)
        try:
            r.load({"timeout": "abc"}, Config)
        except (AggregateLoadError, LoadError) as exc:
            errors = extract_field_errors(exc)
            assert len(errors) == 1
            assert errors[0].field_path == ["timeout"]
            assert errors[0].message == "Bad string format"

    def test_missing_field(self):
        @dataclass
        class Config:
            timeout: int
            name: str

        r = Retort(strict_coercion=False)
        try:
            r.load({"timeout": 123}, Config)
        except (AggregateLoadError, LoadError) as exc:
            errors = extract_field_errors(exc)
            assert len(errors) == 1
            assert errors[0].field_path == ["name"]
            assert errors[0].message == "Missing required field"

    def test_nested_errors(self):
        @dataclass
        class DB:
            host: str
            port: int

        @dataclass
        class Config:
            timeout: int
            db: DB

        r = Retort(strict_coercion=False)
        try:
            r.load({"timeout": "abc", "db": {"host": "ok", "port": "xyz"}}, Config)
        except (AggregateLoadError, LoadError) as exc:
            errors = extract_field_errors(exc)
            assert len(errors) == 2
            paths = [e.field_path for e in errors]
            assert ["timeout"] in paths
            assert ["db", "port"] in paths

    def test_multiple_missing_fields(self):
        @dataclass
        class Config:
            a: int
            b: str
            c: float

        r = Retort(strict_coercion=False)
        try:
            r.load({}, Config)
        except (AggregateLoadError, LoadError) as exc:
            errors = extract_field_errors(exc)
            assert len(errors) == 3
            paths = sorted([e.field_path[0] for e in errors])
            assert paths == ["a", "b", "c"]


class TestResolveSourceLocation:
    def test_env_source(self):
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="env",
            file_path=None,
            prefix="APP_",
            split_symbols="__",
        )
        loc = resolve_source_location(["database", "port"], ctx, file_content=None)
        assert loc.source_type == "env"
        assert loc.env_var_name == "APP_DATABASE__PORT"
        assert loc.file_path is None

    def test_env_source_no_prefix(self):
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="env",
            file_path=None,
            prefix=None,
            split_symbols="__",
        )
        loc = resolve_source_location(["timeout"], ctx, file_content=None)
        assert loc.env_var_name == "TIMEOUT"

    def test_env_source_custom_split_symbols(self):
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="env",
            file_path=None,
            prefix="APP_",
            split_symbols="_",
        )
        loc = resolve_source_location(["database", "port"], ctx, file_content=None)
        assert loc.env_var_name == "APP_DATABASE_PORT"

    def test_json_source_with_line(self):
        content = '{\n  "timeout": "30",\n  "name": "test"\n}'
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="json",
            file_path=Path("config.json"),
            prefix=None,
            split_symbols="__",
        )
        loc = resolve_source_location(["timeout"], ctx, file_content=content)
        assert loc.source_type == "json"
        assert loc.line_number == 2
        assert loc.line_content == '"timeout": "30",'

    def test_toml_source_with_line(self):
        content = 'timeout = "30"\nname = "test"'
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="toml",
            file_path=Path("config.toml"),
            prefix=None,
            split_symbols="__",
        )
        loc = resolve_source_location(["timeout"], ctx, file_content=content)
        assert loc.source_type == "toml"
        assert loc.line_number == 1
        assert loc.line_content == 'timeout = "30"'

    def test_envfile_source(self):
        content = "# comment\nAPP_TIMEOUT=30\nAPP_NAME=test"
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="envfile",
            file_path=Path(".env"),
            prefix="APP_",
            split_symbols="__",
        )
        loc = resolve_source_location(["timeout"], ctx, file_content=content)
        assert loc.source_type == "envfile"
        assert loc.env_var_name == "APP_TIMEOUT"
        assert loc.line_number == 2
        assert loc.line_content == "APP_TIMEOUT=30"


class TestDatureConfigErrorFormat:
    def test_single_error_message(self):
        errors = [
            FieldError(
                error=FieldErrorInfo(
                    field_path=["timeout"],
                    message="Expected int, got str",
                    input_value="30",
                ),
                location=SourceLocation(
                    source_type="toml",
                    file_path=Path("config.toml"),
                    line_number=2,
                    line_content='timeout = "30"',
                    env_var_name=None,
                ),
            ),
        ]
        exc = DatureConfigError(errors, "Config")
        assert str(exc) == dedent("""\
            Config loading errors (1)

              [timeout]  Expected int, got str
               └── FILE 'config.toml', line 2
                   timeout = "30"
            """)

    def test_multiple_errors_message(self):
        errors = [
            FieldError(
                error=FieldErrorInfo(
                    field_path=["timeout"],
                    message="Bad string format",
                    input_value="abc",
                ),
                location=SourceLocation(
                    source_type="json",
                    file_path=Path("config.json"),
                    line_number=2,
                    line_content='"timeout": "abc"',
                    env_var_name=None,
                ),
            ),
            FieldError(
                error=FieldErrorInfo(
                    field_path=["db", "port"],
                    message="Missing required field",
                    input_value=None,
                ),
                location=SourceLocation(
                    source_type="json",
                    file_path=Path("config.json"),
                    line_number=None,
                    line_content=None,
                    env_var_name=None,
                ),
            ),
        ]
        exc = DatureConfigError(errors, "Config")
        assert str(exc) == dedent("""\
            Config loading errors (2)

              [timeout]  Bad string format
               └── FILE 'config.json', line 2
                   "timeout": "abc"

              [db.port]  Missing required field
               └── FILE 'config.json'
            """)

    def test_env_error_message(self):
        errors = [
            FieldError(
                error=FieldErrorInfo(
                    field_path=["database", "port"],
                    message="Bad string format",
                    input_value="abc",
                ),
                location=SourceLocation(
                    source_type="env",
                    file_path=None,
                    line_number=None,
                    line_content=None,
                    env_var_name="APP_DATABASE__PORT",
                ),
            ),
        ]
        exc = DatureConfigError(errors, "Config")
        assert str(exc) == dedent("""\
            Config loading errors (1)

              [database.port]  Bad string format
               └── ENV 'APP_DATABASE__PORT'
            """)


class TestLoadIntegrationErrors:
    def test_json_type_error_decorator(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"timeout": "abc", "name": "test"}')

        metadata = LoadMetadata(file_=str(json_file))

        @load(metadata)
        @dataclass
        class Config:
            timeout: int
            name: str

        with pytest.raises(DatureConfigError) as exc_info:
            Config()

        err = exc_info.value
        assert len(err.errors) == 1
        assert err.errors[0].error.field_path == ["timeout"]
        assert str(json_file) in str(err)

    def test_json_missing_field_function(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test"}')

        @dataclass
        class Config:
            name: str
            port: int

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert len(err.errors) == 1
        assert err.errors[0].error.field_path == ["port"]
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [port]  Missing required field
               └── FILE '{json_file}'
            """)

    def test_multiple_errors_at_once(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"timeout": "abc"}')

        @dataclass
        class Config:
            timeout: int
            name: str

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert len(err.errors) == 2
        paths = [e.error.field_path for e in err.errors]
        assert ["timeout"] in paths
        assert ["name"] in paths
        assert "Config loading errors (2)" in str(err)

    def test_nested_dataclass_error(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{\n  "db": {\n    "host": "localhost",\n    "port": "abc"\n  }\n}')

        @dataclass
        class DB:
            host: str
            port: int

        @dataclass
        class Config:
            db: DB

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert len(err.errors) == 1
        assert err.errors[0].error.field_path == ["db", "port"]
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [db.port]  Bad string format
               └── FILE '{json_file}', line 4
                   "port": "abc"
            """)

    def test_env_type_error(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("APP_TIMEOUT", "abc")
        monkeypatch.setenv("APP_NAME", "test")

        metadata = LoadMetadata(prefix="APP_")

        @load(metadata)
        @dataclass
        class Config:
            timeout: int
            name: str

        with pytest.raises(DatureConfigError) as exc_info:
            Config()

        err = exc_info.value
        assert len(err.errors) == 1
        assert str(err) == dedent("""\
            Config loading errors (1)

              [timeout]  Bad string format
               └── ENV 'APP_TIMEOUT'
            """)

    def test_toml_with_line_number(self, tmp_path: Path):
        toml_file = tmp_path / "config.toml"
        toml_file.write_text('name = "test"\ntimeout = "abc"\n')

        @dataclass
        class Config:
            name: str
            timeout: int

        metadata = LoadMetadata(file_=str(toml_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert len(err.errors) == 1
        assert err.errors[0].location is not None
        assert err.errors[0].location.line_number == 2
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [timeout]  Bad string format
               └── FILE '{toml_file}', line 2
                   timeout = "abc"
            """)

    def test_json_with_line_number(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{\n  "name": "test",\n  "timeout": "abc"\n}')

        @dataclass
        class Config:
            name: str
            timeout: int

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert err.errors[0].location is not None
        assert err.errors[0].location.line_number == 3
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [timeout]  Bad string format
               └── FILE '{json_file}', line 3
                   "timeout": "abc"
            """)



import json
import re
from typing import Any, cast

from dature.sources_loader.loaders.common import bool_from_string
from dature.types import JSONValue


def dict_from_env_nested(value: str | dict[Any, Any]) -> dict[Any, Any]:
    if isinstance(value, str):
        if value == "":
            return {}

        value_fixed = re.sub(r"(\{|,)\s*(\d+)\s*:", r'\1"\2":', value)
        parsed = json.loads(value_fixed)

        if not isinstance(parsed, dict):
            msg = f"Expected dict in JSON, got {type(parsed)}"
            raise TypeError(msg)

        if all(k.isdigit() for k in parsed if isinstance(k, str)):
            return {int(k): v for k, v in parsed.items()}

        return parsed

    if isinstance(value, dict):
        result: dict[Any, Any] = {}
        for k, v in value.items():
            if isinstance(v, dict):
                result[k] = dict_from_env_nested(v)
            elif isinstance(v, str):
                result[k] = _infer_type(v)
            else:
                result[k] = v

        if all(k.isdigit() for k in result if isinstance(k, str)):
            return {int(k): v for k, v in result.items()}

        return result

    return value


def _infer_type(value: str) -> JSONValue:
    if value == "":
        return value

    try:
        return bool_from_string(value)
    except TypeError:
        pass

    if value.startswith(("[", "{")):
        try:
            return cast("JSONValue", json.loads(value))
        except json.JSONDecodeError:
            pass

    try:
        if "." in value:
            return float(value)
        return int(value)
    except ValueError:
        pass

    return value



from pathlib import Path
from types import MappingProxyType
from typing import Any, Literal

from dature.metadata import LoadMetadata
from dature.sources_loader.base import ILoader
from dature.sources_loader.env_ import EnvFileLoader, EnvLoader
from dature.sources_loader.ini_ import IniLoader
from dature.sources_loader.json_ import JsonLoader
from dature.sources_loader.toml_ import TomlLoader

LoaderType = Literal["env", "envfile", "yaml", "yaml1.1", "yaml1.2", "json", "json5", "toml", "ini"]

EXTENSION_MAP: MappingProxyType[str, LoaderType] = MappingProxyType(
    {
        ".env": "envfile",
        ".yaml": "yaml",
        ".yml": "yaml",
        ".json": "json",
        ".json5": "json5",
        ".toml": "toml",
        ".ini": "ini",
        ".cfg": "ini",
    },
)


def _get_loader_class(loader_type: LoaderType) -> type[ILoader]:
    match loader_type:
        case "env":
            return EnvLoader
        case "envfile":
            return EnvFileLoader
        case "yaml" | "yaml1.1":
            from dature.sources_loader.yaml_ import Yaml11Loader  # noqa: PLC0415

            return Yaml11Loader
        case "yaml1.2":
            from dature.sources_loader.yaml_ import Yaml12Loader  # noqa: PLC0415

            return Yaml12Loader
        case "json":
            return JsonLoader
        case "json5":
            from dature.sources_loader.json5_ import Json5Loader  # noqa: PLC0415

            return Json5Loader
        case "toml":
            return TomlLoader
        case "ini":
            return IniLoader
        case _:
            msg = f"Unknown loader type: {loader_type}"
            raise ValueError(msg)


def get_loader_type(metadata: LoadMetadata) -> LoaderType:
    if metadata.loader:
        return metadata.loader

    if not metadata.file_:
        return "env"

    file_path = Path(metadata.file_)

    if (extension := file_path.suffix.lower()) in EXTENSION_MAP:
        return EXTENSION_MAP[extension]

    if file_path.name.startswith(".env"):
        return "envfile"

    supported = ", ".join(EXTENSION_MAP.keys())
    msg = (
        f"Cannot determine loader type for file '{metadata.file_}'. "
        f"Please specify loader explicitly or use a supported extension: {supported}"
    )
    raise ValueError(msg)


def resolve_loader(metadata: LoadMetadata) -> ILoader:
    loader_type = get_loader_type(metadata)
    loader_class = _get_loader_class(loader_type)

    kwargs: dict[str, Any] = {
        "prefix": metadata.prefix,
        "name_style": metadata.name_style,
        "field_mapping": metadata.field_mapping,
        "root_validators": metadata.root_validators,
        "enable_expand_env_vars": metadata.enable_expand_env_vars,
    }

    if issubclass(loader_class, EnvLoader):
        kwargs["split_symbols"] = metadata.split_symbols

    return loader_class(**kwargs)



"""Tests for FieldPath lazy field path builder."""

from dataclasses import dataclass

import pytest

from dature.field_path import F, FieldPath


@dataclass
class _Database:
    uri: str


@dataclass
class _Cfg:
    host: str
    database: _Database


class TestFieldPath:
    @pytest.mark.parametrize(
        ("field_path", "expected_path"),
        [
            pytest.param(F[_Cfg].host, "host", id="single_field_class"),
            pytest.param(F[_Cfg].database.uri, "database.uri", id="nested_field_class"),
            pytest.param(F["Config"].host, "host", id="single_field_string"),
            pytest.param(F["Config"].database.uri, "database.uri", id="nested_field_string"),
        ],
    )
    def test_path(self, field_path: FieldPath, expected_path: str):
        assert field_path.as_path() == expected_path

    def test_owner_is_class(self):
        assert F[_Cfg].host.owner is _Cfg

    def test_owner_is_string(self):
        assert F["Config"].host.owner == "Config"

    def test_no_fields_raises_value_error(self):
        with pytest.raises(ValueError, match="at least one field name"):
            F[_Cfg].as_path()

    def test_is_frozen(self):
        fp = F[_Cfg].host
        with pytest.raises(AttributeError):
            fp.owner = "other"

    def test_nonexistent_field_raises_attribute_error(self):
        with pytest.raises(AttributeError, match="'_Cfg' has no field 'nonexistent'"):
            _ = F[_Cfg].nonexistent

    def test_not_a_dataclass_raises_type_error(self):
        class Plain:
            pass

        with pytest.raises(TypeError, match="'Plain' is not a dataclass"):
            F[Plain]

    def test_string_owner_skips_validation(self):
        fp = F["Whatever"].anything.deep.path
        assert fp.as_path() == "anything.deep.path"



from dataclasses import dataclass


class JsonPathFinder:
    def __init__(self, content: str) -> None:
        self._content = content

    def find_line(self, target_path: list[str]) -> int:
        content = self._content
        length = len(content)
        pos = 0
        line = 1
        stack: list[_StackEntry] = []
        last_key: str | None = None

        while pos < length:
            ch = content[pos]

            if ch in " \t\r\n":
                if ch == "\n":
                    line += 1
                pos += 1
                continue

            if ch == ",":
                pos += 1
                continue

            if ch in "{[":
                _increment_parent_array(stack)
                is_array = ch == "["
                stack.append(_StackEntry(key=last_key, is_array=is_array, array_index=-1))
                pos += 1
                last_key = None
                continue

            if ch in "}]":
                if stack:
                    stack.pop()
                pos += 1
                last_key = None
                continue

            if ch != '"':
                _increment_parent_array(stack)
                pos = _skip_scalar(content, pos, length)
                continue

            result = _handle_string(content, pos, length, stack, target_path)
            pos = result.pos
            last_key = result.last_key
            if result.found:
                return line

        return -1


@dataclass
class _StackEntry:
    key: str | None
    is_array: bool
    array_index: int


@dataclass(frozen=True, slots=True)
class _StringResult:
    pos: int
    last_key: str | None
    found: bool


def _handle_string(
    content: str,
    pos: int,
    length: int,
    stack: list[_StackEntry],
    target_path: list[str],
) -> _StringResult:
    string_start = pos + 1
    end_pos = _skip_string(content, string_start)
    string_value = content[string_start : end_pos - 1]

    colon_pos = _skip_ws(content, end_pos, length)

    if colon_pos >= length or content[colon_pos] != ":":
        _increment_parent_array(stack)
        return _StringResult(pos=end_pos, last_key=None, found=False)

    if _build_path(stack, string_value) == target_path:
        return _StringResult(pos=end_pos, last_key=None, found=True)

    return _StringResult(pos=colon_pos + 1, last_key=string_value, found=False)


def _increment_parent_array(stack: list[_StackEntry]) -> None:
    if stack and stack[-1].is_array:
        stack[-1].array_index += 1


def _skip_string(content: str, pos: int) -> int:
    """Advance past a JSON string body. Returns position after the closing quote."""
    length = len(content)
    while pos < length:
        ch = content[pos]
        if ch == "\\":
            pos += 2
            continue
        if ch == '"':
            return pos + 1
        pos += 1
    return pos


def _skip_ws(content: str, pos: int, length: int) -> int:
    while pos < length and content[pos] in " \t\r\n":
        pos += 1
    return pos


def _skip_scalar(content: str, pos: int, length: int) -> int:
    while pos < length and content[pos] not in " \t\r\n,}]":
        pos += 1
    return pos


def _build_path(stack: list[_StackEntry], current_key: str) -> list[str]:
    path: list[str] = []
    for entry in stack:
        if entry.key is not None:
            path.append(entry.key)
        if entry.is_array:
            path.append(str(entry.array_index))
    path.append(current_key)
    return path



import json
import re
from typing import Any


def dict_from_json_string(value: str | dict[Any, Any]) -> dict[Any, Any]:
    if isinstance(value, dict):
        return value

    if value == "":
        return {}

    value_fixed = re.sub(r"(\{|,)\s*(\d+)\s*:", r'\1"\2":', value)
    parsed = json.loads(value_fixed)

    if not isinstance(parsed, dict):
        msg = f"Expected dict in JSON, got {type(parsed)}"
        raise TypeError(msg)

    if all(k.isdigit() for k in parsed if isinstance(k, str)):
        return {int(k): v for k, v in parsed.items()}

    return parsed



from typing import TYPE_CHECKING

from dature.field_path import FieldPath

if TYPE_CHECKING:
    from dature.metadata import FieldMergeStrategy, MergeRule


def extract_field_path(predicate: object) -> str:
    if not isinstance(predicate, FieldPath):
        msg = f"Expected FieldPath, got {type(predicate).__name__}"
        raise TypeError(msg)
    return predicate.as_path()


def build_field_merge_map(field_merges: "tuple[MergeRule, ...]") -> "dict[str, FieldMergeStrategy]":
    result: dict[str, FieldMergeStrategy] = {}
    for rule in field_merges:
        path = extract_field_path(rule.predicate)
        result[path] = rule.strategy
    return result



from collections.abc import Callable
from dataclasses import Field
from typing import Any, ClassVar, Protocol


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


class ValidatorProtocol(Protocol):
    def get_validator_func(self) -> Callable[..., bool]: ...

    def get_error_message(self) -> str: ...



from typing import Annotated, get_args, get_origin

from adaptix import P, validator
from adaptix.provider import Provider

from dature.validators.protocols import ValidatorProtocol


def extract_validators_from_type(field_type: object) -> list[ValidatorProtocol]:
    validators: list[ValidatorProtocol] = []

    if get_origin(field_type) is not Annotated:
        return validators

    args = get_args(field_type)

    validators.extend(arg for arg in args[1:] if hasattr(arg, "__dataclass_fields__"))

    return validators


def create_validator_providers(
    dataclass_: type,
    field_name: str,
    validators: list[ValidatorProtocol],
) -> list[Provider]:
    providers = []

    for v in validators:
        func = v.get_validator_func()
        error = v.get_error_message()
        provider = validator(
            P[dataclass_][field_name],
            func,
            error,
        )
        providers.append(provider)

    return providers


def create_root_validator_providers(
    dataclass_: type,
    root_validators: tuple[ValidatorProtocol, ...],
) -> list[Provider]:
    providers = []

    for root_validator in root_validators:
        provider = validator(
            P[dataclass_],
            root_validator.get_validator_func(),
            root_validator.get_error_message(),
        )
        providers.append(provider)

    return providers



import types
from typing import Annotated, Literal
from urllib.parse import ParseResult

type JSONValue = dict[str, JSONValue] | list[JSONValue] | str | int | float | bool | None

# Result of get_type_hints() / get_args(): concrete class or parameterized generic
type TypeAnnotation = type[object] | types.GenericAlias

# Examples: "app", "app.database", "app.database.host"
type DotSeparatedPath = Annotated[str, "Dot-separated path for nested dictionary navigation"]

type NameStyle = Literal[
    "lower_snake",
    "upper_snake",
    "lower_camel",
    "upper_camel",
    "lower_kebab",
    "upper_kebab",
]

type FieldMapping = dict[str, str]

type URL = ParseResult

type Base64UrlBytes = bytes
type Base64UrlStr = str



from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from typing import Annotated, Literal

import pytest

from dature import LoadMetadata, load
from dature.errors import DatureConfigError
from dature.validators.number import Ge, Le
from dature.validators.sequence import MinItems, UniqueItems
from dature.validators.string import MaxLength, MinLength, RegexPattern


@dataclass
class Address:
    city: Annotated[str, MinLength(value=2)]
    zip_code: Annotated[str, RegexPattern(pattern=r"^\d{5}$")]


@dataclass
class ErrorConfig:
    port: int
    host: str
    status: Literal["active", "inactive"]
    name: Annotated[str, MinLength(value=3), MaxLength(value=50)]
    email: Annotated[str, RegexPattern(pattern=r"^[\w.-]+@[\w.-]+\.\w+$")]
    age: Annotated[int, Ge(value=0), Le(value=150)]
    tags: Annotated[list[str], MinItems(value=1), UniqueItems()]
    address: Address


# adaptix stops at load errors before running validators,
# so we split into two dataclasses to test both phases separately
@dataclass
class LoadErrorConfig:
    port: int
    host: str
    status: Literal["active", "inactive"]
    name: str
    email: str
    age: int
    tags: list[str]
    address: Address


@dataclass
class ValidationErrorConfig:
    name: Annotated[str, MinLength(value=3), MaxLength(value=50)]
    email: Annotated[str, RegexPattern(pattern=r"^[\w.-]+@[\w.-]+\.\w+$")]
    age: Annotated[int, Ge(value=0), Le(value=150)]
    tags: Annotated[list[str], MinItems(value=1), UniqueItems()]
    address: Address


FIXTURES_DIR = Path(__file__).parent / "fixtures"

ALL_SOURCES = [
    ("errors.json", {}),
    ("errors.json5", {}),
    ("errors.yaml", {}),
    ("errors.yaml", {"loader": "yaml1.1"}),
    ("errors.yaml", {"loader": "yaml1.2"}),
    ("errors.toml", {}),
    ("errors.ini", {"prefix": "config"}),
    ("errors.env", {}),
]


def _load_error_message(p: str) -> str:
    return dedent("""\
        LoadErrorConfig loading errors (3)

          [port]  Bad string format
           └── FILE '{path}', line 2
               "port": "abc",

          [host]  Missing required field
           └── FILE '{path}'

          [status]  Invalid variant: 'unknown'
           └── FILE '{path}', line 3
               "status": "unknown",
        """).replace("{path}", p)


def _validation_error_message(p: str) -> str:
    return dedent("""\
        ValidationErrorConfig loading errors (6)

          [name]  Value must have at least 3 characters
           └── FILE '{path}', line 4
               "name": "AB",

          [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
           └── FILE '{path}', line 5
               "email": "not-an-email",

          [age]  Value must be less than or equal to 150
           └── FILE '{path}', line 6
               "age": 200,

          [tags]  Value must have at least 1 items
           └── FILE '{path}', line 7
               "tags": [],

          [address.city]  Value must have at least 2 characters
           └── FILE '{path}', line 9
               "city": "N",

          [address.zip_code]  Value must match pattern '^\\d{5}$'
           └── FILE '{path}', line 10
               "zip_code": "ABCDE"
        """).replace("{path}", p)


EXPECTED_LOAD_MESSAGES: dict[str, str] = {}
EXPECTED_VALIDATION_MESSAGES: dict[str, str] = {}


def _file(name: str) -> str:
    return str(FIXTURES_DIR / name)


# --- json ---
EXPECTED_LOAD_MESSAGES["json"] = _load_error_message(_file("errors.json"))
EXPECTED_VALIDATION_MESSAGES["json"] = _validation_error_message(_file("errors.json"))

# --- json5 ---
_json5_path = _file("errors.json5")
EXPECTED_LOAD_MESSAGES["json5"] = dedent(f"""\
    LoadErrorConfig loading errors (3)

      [port]  Bad string format
       └── FILE '{_json5_path}', line 3
           port: "abc",

      [host]  Missing required field
       └── FILE '{_json5_path}'

      [status]  Invalid variant: 'unknown'
       └── FILE '{_json5_path}', line 5
           status: "unknown",
    """)
EXPECTED_VALIDATION_MESSAGES["json5"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── FILE '{_json5_path}', line 7
           name: "AB",

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── FILE '{_json5_path}', line 9
           email: "not-an-email",

      [age]  Value must be less than or equal to 150
       └── FILE '{_json5_path}', line 11
           age: 200,

      [tags]  Value must have at least 1 items
       └── FILE '{_json5_path}', line 13
           tags: [],

      [address.city]  Value must have at least 2 characters
       └── FILE '{_json5_path}', line 16
           city: "N",

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_json5_path}', line 17
           zip_code: "ABCDE",
    """)

# --- yaml ---
_yaml_path = _file("errors.yaml")
EXPECTED_LOAD_MESSAGES["yaml"] = dedent(f"""\
    LoadErrorConfig loading errors (3)

      [port]  Bad string format
       └── FILE '{_yaml_path}', line 1
           port: "abc"

      [host]  Missing required field
       └── FILE '{_yaml_path}'

      [status]  Invalid variant: 'unknown'
       └── FILE '{_yaml_path}', line 2
           status: "unknown"
    """)
EXPECTED_VALIDATION_MESSAGES["yaml"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── FILE '{_yaml_path}', line 3
           name: "AB"

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── FILE '{_yaml_path}', line 4
           email: "not-an-email"

      [age]  Value must be less than or equal to 150
       └── FILE '{_yaml_path}', line 5
           age: 200

      [tags]  Value must have at least 1 items
       └── FILE '{_yaml_path}', line 6
           tags: []

      [address.city]  Value must have at least 2 characters
       └── FILE '{_yaml_path}', line 8
           city: "N"

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_yaml_path}', line 9
           zip_code: "ABCDE"
    """)

# --- toml ---
_toml_path = _file("errors.toml")
EXPECTED_LOAD_MESSAGES["toml"] = dedent(f"""\
    LoadErrorConfig loading errors (3)

      [port]  Bad string format
       └── FILE '{_toml_path}', line 1
           port = "abc"

      [host]  Missing required field
       └── FILE '{_toml_path}'

      [status]  Invalid variant: 'unknown'
       └── FILE '{_toml_path}', line 2
           status = "unknown"
    """)
EXPECTED_VALIDATION_MESSAGES["toml"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── FILE '{_toml_path}', line 3
           name = "AB"

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── FILE '{_toml_path}', line 4
           email = "not-an-email"

      [age]  Value must be less than or equal to 150
       └── FILE '{_toml_path}', line 5
           age = 200

      [tags]  Value must have at least 1 items
       └── FILE '{_toml_path}', line 6
           tags = []

      [address.city]  Value must have at least 2 characters
       └── FILE '{_toml_path}', line 9
           city = "N"

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_toml_path}', line 10
           zip_code = "ABCDE"
    """)

# --- ini ---
_ini_path = _file("errors.ini")
EXPECTED_LOAD_MESSAGES["ini"] = dedent(f"""\
    LoadErrorConfig loading errors (3)

      [port]  Bad string format
       └── FILE '{_ini_path}', line 2
           port = abc

      [host]  Missing required field
       └── FILE '{_ini_path}'

      [status]  Invalid variant: 'unknown'
       └── FILE '{_ini_path}', line 3
           status = unknown
    """)
EXPECTED_VALIDATION_MESSAGES["ini"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── FILE '{_ini_path}', line 4
           name = AB

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── FILE '{_ini_path}', line 5
           email = not-an-email

      [age]  Value must be less than or equal to 150
       └── FILE '{_ini_path}', line 6
           age = 200

      [tags]  Value must have at least 1 items
       └── FILE '{_ini_path}', line 7
           tags = []

      [address.city]  Value must have at least 2 characters
       └── FILE '{_ini_path}', line 10
           city = N

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_ini_path}', line 11
           zip_code = ABCDE
    """)

# --- env ---
_env_path = _file("errors.env")
EXPECTED_LOAD_MESSAGES["env"] = dedent(f"""\
    LoadErrorConfig loading errors (3)

      [port]  Bad string format
       └── ENV FILE '{_env_path}', var 'PORT'
           PORT=abc

      [host]  Missing required field
       └── ENV FILE '{_env_path}', var 'HOST'

      [status]  Invalid variant: 'unknown'
       └── ENV FILE '{_env_path}', var 'STATUS'
           STATUS=unknown
    """)
EXPECTED_VALIDATION_MESSAGES["env"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── ENV FILE '{_env_path}', var 'NAME'
           NAME=AB

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── ENV FILE '{_env_path}', var 'EMAIL'
           EMAIL=not-an-email

      [age]  Value must be less than or equal to 150
       └── ENV FILE '{_env_path}', var 'AGE'
           AGE=200

      [tags]  Value must have at least 1 items
       └── ENV FILE '{_env_path}', var 'TAGS'
           TAGS=[]

      [address.city]  Value must have at least 2 characters
       └── ENV FILE '{_env_path}', var 'ADDRESS__CITY'
           ADDRESS__CITY=N

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── ENV FILE '{_env_path}', var 'ADDRESS__ZIP_CODE'
           ADDRESS__ZIP_CODE=ABCDE
    """)


def _get_format_key(fixture_file: str) -> str:
    suffix = Path(fixture_file).suffix.lstrip(".")
    if suffix == "json5":
        return "json5"
    return suffix


@pytest.mark.parametrize(("fixture_file", "metadata_kwargs"), ALL_SOURCES)
def test_load_error_types(
    fixture_file: str,
    metadata_kwargs: dict[str, str],
) -> None:
    path = str(FIXTURES_DIR / fixture_file)
    metadata = LoadMetadata(file_=path, **metadata_kwargs)

    with pytest.raises(DatureConfigError) as exc_info:
        load(metadata, LoadErrorConfig)

    fmt = _get_format_key(fixture_file)
    assert str(exc_info.value) == EXPECTED_LOAD_MESSAGES[fmt]


@pytest.mark.parametrize(("fixture_file", "metadata_kwargs"), ALL_SOURCES)
def test_validation_error_types(
    fixture_file: str,
    metadata_kwargs: dict[str, str],
) -> None:
    path = str(FIXTURES_DIR / fixture_file)
    metadata = LoadMetadata(file_=path, **metadata_kwargs)

    with pytest.raises(DatureConfigError) as exc_info:
        load(metadata, ValidationErrorConfig)

    fmt = _get_format_key(fixture_file)
    assert str(exc_info.value) == EXPECTED_VALIDATION_MESSAGES[fmt]
